# =====================================
# generator=datazen
# version=3.1.4
# hash=bf2c126f5c469e9b0483e934ee695f9e
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "3.3.0"
